/*
 * $Id: become.c,v 1.1.1.1 2006/03/21 23:44:12 dave Exp $
 */
#include "become.h"

char	*args[] = {"-", NULL};
char	*vars[] = {"PATH", "TERM", NULL};
char	**env = NULL;
int	env_count = 0;

int	verbosity = 0;
int	realids = 0;
int	clearsup = 0;

void
usage (char *name) {
	fprintf(stderr, "Usage: %s [options] [uid][:gid]\n", name);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -h   Display usage information\n");
	fprintf(stderr, "  -v   Increase verbosity\n");
	fprintf(stderr, "  -c   Clear supplementary groups\n");
	fprintf(stderr, "  -r   Set real ids instead of effective ids\n");
}

int
main (int argc, char *argv[]) {
	char		*cp;
	int		i;
	long		old_uid, new_uid, old_gid, new_gid;
	pid_t		pid;

	while ((i = getopt(argc, argv, "hvcr")) != -1) {
		switch(i) {
		case 'h':
			usage(argv[0]);
			exit(EX_OK);
		case 'v':
			verbosity++;
			break;
		case 'c':
			clearsup = 1;
			break;
		case 'r':
			realids = 1;
			break;
		case '?':
			exit(EX_USAGE);
		}
	}

	argc -= optind;
	argv += optind;
	if (argc < 1)
		errx(EX_USAGE, "No user or group specified!");

	if (realids != 0) {
		new_uid = old_uid = getuid();
		new_gid = old_gid = getgid();
	} else {
		new_uid = old_uid = geteuid();
		new_gid = old_gid = getegid();
	}

	if ((cp = strchr(argv[0], ':')) != 0) {
		*cp++ = 0x00;
	} else
		cp = strchr(argv[0], 0x00);
	
	if (strlen(argv[0]) > 0) {
		new_uid = strtol(argv[0], NULL, 10);
		if (errno == EINVAL)
			errx(EX_USAGE, "Invalid uid");
		if (new_uid < 0 || new_uid > 65535)
			err(EX_USAGE, "uid out of range");
	}
	if (strlen(cp) > 0) {
		new_gid = strtol(cp, NULL, 10);
		if (errno == EINVAL)
			errx(EX_USAGE, "Invalid gid");
		if (new_gid < 0 || new_gid > 65535)
			err(EX_USAGE, "gid out of range");
	}

	for (i = 0; vars[i] != NULL; i++) {
		if ((cp = getenv(vars[i])) == NULL)
			continue;
		if ((env = realloc(env, (env_count+1) * sizeof(char *))) == NULL)
			err(EX_OSERR, "realloc() failed");
		cp -= strlen(vars[i])+1;
		env[env_count++] = cp;
	}

	if (new_gid != old_gid) {
		if (realids != 0) {
			if (setgid(new_gid) < 0)
				err(EX_OSERR, "setgid() failed");
		} else {
			if (setegid(new_gid) < 0)
				err(EX_OSERR, "setegid() failed");
		}

		if (clearsup != 0)
			if (setgroups(0, NULL) != 0)
				err(EX_OSERR, "setgroups() failed");
	}
	if (new_uid != old_uid) {
		if (realids != 0) {
			if (setuid(new_uid) < 0)
				err(EX_OSERR, "setuid() failed");
		} else {
			if (seteuid(new_uid) < 0)
				err(EX_OSERR, "seteuid() failed");
		}
	}

	printf("euid=%d, uid=%d, egid=%d, gid=%d\n",
	    geteuid(), getuid(), getegid(), getgid());

	if ((pid = fork()) == 0) {
		execve("/bin/sh", args, env);
		err(EX_OSERR, "execve() failed");
	}

	if (pid < 0)
		err(EX_OSERR, "fork() failed");

	if (waitpid(pid, NULL, 0) != pid)
		warnx("waitpid() returned someone else child!");

	return(EX_OK);
}
